package net.pricefx.domain;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import org.joda.time.DateTime;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import java.lang.reflect.Field;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Supplier;

import static java.lang.annotation.RetentionPolicy.RUNTIME;

/**
 * A few important constraints for persisted objects:
 * <p><ul>
 *     <li>MUST NOT be final, neither class nor persisted fields, see <a href="https://stackoverflow.com/a/47771587">this</a></li>
 *     <li>JSON persisted fields should implement equals/hashCode, see <a href="https://github.com/vladmihalcea/hibernate-types#best-practices">this</a></li>
 * </ul><p>
 */
public abstract class PersistedObject {

    public static final String FIELD_ID = "id";
    public static final String FIELD_LASTUPDATEDATE = "lastUpdateDate";
	public static final String FIELD_CREATEDATE = "createDate";
	public static final String FIELD_CREATEDBY = "createdBy";
	public static final String FIELD_ISDELETED = "isDeleted";

	public PersistedObject() {
		LocalDateTime now = LocalDateTime.now();
//		Long currentUser = UserContext.getCurrentUser();
		this.createDate = now;
//		this.createdBy = currentUser;
		this.lastUpdateDate = now;
//		this.lastUpdateBy = currentUser;
	}

	private Long id;

	private Long version;

	@JsonIgnore
    private LocalDateTime createDate;

    private Long createdBy;

    private LocalDateTime lastUpdateDate;

    private Long lastUpdateBy;

    private LocalDateTime deleteDate;

    private Long deletedBy;

    private boolean isDeleted;

    private String typedId;

	@JsonProperty(value="createdBy")
    public Long getCreatedBy() {
		return createdBy;
	}

	@JsonIgnore
	public void setCreatedBy(Long createdBy) {
		this.createdBy = createdBy;
	}

	@JsonProperty(value="lastUpdateBy")
	public Long getLastUpdateBy() {
		return lastUpdateBy;
	}

	@JsonIgnore
	public void setLastUpdateBy(Long lastUpdateBy) {
		this.lastUpdateBy = lastUpdateBy;
	}

	@JsonIgnore
	public LocalDateTime getDeleteDate() {
		return deleteDate;
	}

	@JsonIgnore
	public void setDeleteDate(LocalDateTime deleteDate) {
		this.deleteDate = deleteDate;
	}

	@JsonIgnore
	public Long getDeletedBy() {
		return deletedBy;
	}

	@JsonIgnore
	public void setDeletedBy(Long deletedBy) {
		this.deletedBy = deletedBy;
	}

	@JsonIgnore
	public boolean isDeleted() {
		return isDeleted;
	}

	@JsonIgnore
	public void setDeleted(boolean isDeleted) {
		this.isDeleted = isDeleted;
	}

	public Long getVersion() {
		return version;
	}

	public void setVersion(Long version) {
		this.version = version;
	}

	public void bumpVersion() {
		this.version = Long.valueOf(this.version.longValue() + 1);
	}

	@JsonIgnore
	public LocalDateTime getCreateDate() {
		return createDate;
	}

/*
	@JsonProperty(value="createDate")
	public String createDate() {
		if(createDate == null) return null;
		return DateTimeFormatter.ofPattern(S.getDateTimePattern()).format(createDate);	//just to externalize it
	}
*/
	/**
	 * Do not use it directly! Needs to be used for JSON deserialization.
	 * When it wasn't used deserialized object has createdDate set to now.
	 * @param createDate
	 */
	@JsonProperty(value=FIELD_CREATEDATE)
	public void setCreateDateShouldBeIgnored(LocalDateTime createDate) {
		this.createDate = createDate;
	}

	// just to satisfy jackson
	public void setCreateDateTime(DateTime datetime) {}

	@JsonIgnore
	public void setCreateDate(LocalDateTime createDate) {
		this.createDate = createDate;
	}

	@JsonIgnore
	public LocalDateTime getLastUpdateDate() {
		return lastUpdateDate;
	}

	@JsonProperty(value = FIELD_LASTUPDATEDATE)
/*
	public String lastUpdateDate() {
		if(lastUpdateDate == null) return null;
		return (new SimpleDateFormat(S.getDateTimePattern())).format(lastUpdateDate);	//just to externalize it
	}
*/

	// just to satisfy jackson
	public void setLastUpdateDateTime(DateTime dateTime) {}

	@JsonIgnore
	public void setLastUpdateDate(LocalDateTime lastUpdateDate) {
		this.lastUpdateDate = lastUpdateDate;
	}

	@JsonIgnore
	public Long getId() {
		return id;
	}

	@JsonIgnore
	public void setId(Long id) {
		this.id = id;
	}

/*
	public String getTypedId(){
		return (getId() == null)? typedId : (getId() + "." + (getTypeCode() == null ? null : getTypeCode().getCodeString()));
	}
*/

	public void setTypedId( String typedId ){
		this.typedId = typedId;
	}
	
	@JsonIgnore
	public String[] getSignificantFields() {
		return new String[] { FIELD_ID };
	}

	@JsonIgnore
	public LinkedHashMap<String,Object> getBulkloadMap() {
		LinkedHashMap<String,Object> map = new LinkedHashMap<String,Object>();

		//we do not add any fields here as the bulk load command auto-fill these values

		return map;
	}

	/**
	 * This method makes use of the getSignificantFields method to return the actual field values as map by reflection
	 * Only override if you need that extra bit of performance
	 * @return
	 */
	@JsonIgnore
	public Map<String,Object> getBusinessKey() {
		Map<String,Object> busKey = new HashMap<>();
		for (String s : getSignificantFields()) {
			try {
//				Field f = ReflectionUtil.findField(this.getClass(), s);
//				f.setAccessible(true);
//				Object val = f.get(this);
//				busKey.put(s,val);
			} catch(Exception e) {
				busKey.put(s,null);
			}
		}
		return busKey;
	}
	/**
	 * Override to define if an object instance should be really deleted (vs just marked as deleted)
	 * Default: soft-delete
	 */
	public boolean doHardDelete(){
		return false;
	}

	/**
	 * Override this method in case a custom SQL insert statement that takes care of key collisions (Mysql: ON DUPLICATE KEY)
	 * Default: false
	 */
	public boolean hasSafeInsertConfigured(){
		return false;
	}

	/**
	 * Override this to enable audit calls on this object
	 * @return
	 */
	@JsonIgnore
	public boolean isAudited() {
		return false;
	}

	/**
	 * Override to enable change callback
	 *
	 */
	@JsonIgnore
	public boolean notifyOnBusinessKeyChange() {
		return false;
	}


	protected interface BulkloadFilterCheckCallback {
		boolean check(String key, List<UserGroup> userGroups, Map<String, Object> cache);
	}

	@Retention( RetentionPolicy.RUNTIME )
	@Target(ElementType.TYPE)
	public @interface PersistedObjectMeta {
		public String typeCode();

		public boolean fetchPermissionRequired() default true;

		public boolean addPermissionRequired() default true;

		public boolean updatePermissionRequired() default true;

		public boolean deletePermissionRequired() default true;

		public boolean folderRestricted() default false;

		public boolean massEditable() default false;

		public boolean bulkLoadable() default false;

		public boolean canBeGlobal() default false;

		public boolean includeInApi() default false;

		public String name() default "";

		public String[] defaultSortKey() default FIELD_ID;

	}

	@Retention( RetentionPolicy.RUNTIME )
	@Target(ElementType.TYPE)
	public @interface VirtualSortFields {
		public VirtualFieldMapping[] fieldMappings() default { };
	}

	@Target({ })
	@Retention(RUNTIME)
	public @interface VirtualFieldMapping {
		String virtual();
		String field();
		String valuePreffix() default "";
		String valueSuffix() default "";
	}

	@Retention( RetentionPolicy.RUNTIME )
	@Target(ElementType.TYPE)
	public @interface VirtualSearchFields {
		public VirtualFieldMapping[] fieldMappings() default { };
	}

	@Target({ })
	@Retention(RUNTIME)
	public @interface ComputedFieldMapping {
		String computed();
		String[] from();
	}

	@Retention( RetentionPolicy.RUNTIME )
	@Target(ElementType.TYPE)
	public @interface ComputedFields {
		public ComputedFieldMapping[] fieldMappings() default { };
	}


	@Retention( RetentionPolicy.RUNTIME )
	@Target(ElementType.TYPE)
	public @interface BatchDelete {
		public boolean enable() default true;
	}

	public static Map<String, VirtualField> getVirtualFieldMappings(Class<?> clazz) {
		Map<String,VirtualField> fieldTranslation = null;
		VirtualSearchFields vsf = clazz.getAnnotation(VirtualSearchFields.class);
		if(vsf != null) {
			VirtualFieldMapping[] fieldMappings = vsf.fieldMappings();
			fieldTranslation = new HashMap<String, VirtualField>();
			if(fieldMappings != null) {
				for(VirtualFieldMapping m : fieldMappings) {
					fieldTranslation.put(m.virtual(), new VirtualField(m.field(), m.valuePreffix(), m.valueSuffix()));
				}
			}
		}
		return fieldTranslation;
	}

	public static boolean isVirtualField(Class<?> clazz, String fieldName) {
		Map<String, VirtualField> fMappings = PersistedObject.getVirtualFieldMappings(clazz);
		if(fMappings != null && fMappings.containsKey(fieldName)) {
			return true;
		}
		return false;
	}


	public static class VirtualField {

		private String field;
		private String valuePreffix;
		private String valueSuffix;

		public VirtualField(String field, String valuePreffix, String valueSuffix) {
			this.field = field;
			this.valuePreffix = valuePreffix;
			this.valueSuffix = valueSuffix;
		}

		public String getField() {
			return field;
		}

		public String getValuePreffix() {
			return valuePreffix;
		}

		public String getValueSuffix() {
			return valueSuffix;
		}
	}
}
