package net.pricefx.formulaengine.scripting.inputbuilder;

import net.pricefx.common.api.InputType;

import java.util.List;
import java.util.Map;

/**
 * Builder class for input type InputMatrix {@link InputType#INPUTMATRIX}
 */
public class InputMatrixInputBuilder extends AbstractInputBuilder<InputMatrixInputBuilder, Object> {

    public InputMatrixInputBuilder(InputBuilderContext builderContext, String paramName) {
        super(builderContext, InputType.INPUTMATRIX, paramName);
    }

    /**
     * Adds specified columns to the input.
     * <p>Example:
     * <blockquote><pre>
     * api.inputBuilderFactory()
     *      .createInputMatrix("input")
     *      .setColumns(["column1", "column2"])
     *      .setFitFieldWidths(true)
     *      .setDefaultHeight(5)
     *      .setColumnValueOptions([column1: ["option1", "option2"]])
     *      .setNoCellRefresh(true)
     *      .setFixTableHeight(true)
     *      .buildContextParameter()
     * </pre></blockquote>
     *
     * @param columns A list of strings representing names of InputMatrix columns.
     * @return self
     */
    public InputMatrixInputBuilder setColumns(List<String> columns) {
        cp.addParameterConfigEntry("columns", columns);
        return this;
    }

    /**
     * Marks specified columns as read-only. A user will not be able to change values in these columns.
     * <p>Example:
     * <blockquote><pre>
     * api.inputBuilderFactory()
     *      .createInputMatrix("input")
     *      .setColumns(["column1", "column2"])
     *      .setReadOnlyColumns(["column2"])
     *      .setColumnValueOptions([column1: ["option1", "option2"], column2: ["option3", "option4"]])
     *      .setNoCellRefresh(true)
     *      .setFixTableHeight(true)
     *      .buildContextParameter()
     * </pre></blockquote>
     *
     * @param readOnlyColumns A list of strings that specifies columns to be marked as read-only.
     * @return self
     */
    public InputMatrixInputBuilder setReadOnlyColumns(List<String> readOnlyColumns) {
        cp.addParameterConfigEntry("readOnlyColumns", readOnlyColumns);
        return this;
    }

    /**
     * Set fit field widths for the configuration parameters.
     * <p>Example:
     * <blockquote><pre>
     * api.inputBuilderFactory()
     *      .createInputMatrix("input")
     *      .setColumns(["column1", "column2"])
     *      .setReadOnlyColumns(["readOnly1", "readOnly2"])
     *      .setFitFieldWidths(true)
     *      .setDefaultHeight(5)
     *      .setColumnValueOptions([column1: ["option1", "option2"]])
     *      .setCanModifyRows(false)
     *      .setNoCellRefresh(true)
     *      .setFixTableHeight(true)
     *      .buildContextParameter()
     * </pre></blockquote>
     *
     * @param value true to enable, false to disable fit field width
     * @return self
     */
    public InputMatrixInputBuilder setFitFieldWidths(boolean value) {
        cp.setConfigParameter("fitFieldWidths", value);
        return this;
    }

    /**
     * Set default height for the configuration parameters.
     * <p>Example:
     * <blockquote><pre>
     * api.inputBuilderFactory()
     *      .createInputMatrix("input")
     *      .setColumns(["column1", "column2"])
     *      .setReadOnlyColumns(["readOnly1", "readOnly2"])
     *      .setFitFieldWidths(true)
     *      .setDefaultHeight(5)
     *      .setColumnValueOptions([column1: ["option1", "option2"]])
     *      .setCanModifyRows(false)
     *      .setNoCellRefresh(true)
     *      .setFixTableHeight(true)
     *      .buildContextParameter()
     * </pre></blockquote>
     *
     * @param value default height
     * @return self
     */
    public InputMatrixInputBuilder setDefaultHeight(int value) {
        cp.setConfigParameter("defaultHeight", value);
        return this;
    }

    /**
     * Set column value options for the configuration parameters.
     * <p>Example:
     * <blockquote><pre>
     * api.inputBuilderFactory()
     *      .createInputMatrix("input")
     *      .setColumns(["column1", "column2"])
     *      .setReadOnlyColumns(["readOnly1", "readOnly2"])
     *      .setFitFieldWidths(true)
     *      .setDefaultHeight(5)
     *      .setColumnValueOptions([column1: ["option1", "option2"]])
     *      .setCanModifyRows(false)
     *      .setNoCellRefresh(true)
     *      .setFixTableHeight(true)
     *      .buildContextParameter()
     * </pre></blockquote>
     *
     * @param value Map of the column value options
     * @return self
     */
    public InputMatrixInputBuilder setColumnValueOptions(Map value) {
        cp.setConfigParameter("columnValueOptions", value);
        return this;
    }

    /**
     * Set row modification flag of the configuration parameters.
     * <p>Example:
     * <blockquote><pre>
     * api.inputBuilderFactory()
     *      .createInputMatrix("input")
     *      .setColumns(["column1", "column2"])
     *      .setReadOnlyColumns(["readOnly1", "readOnly2"])
     *      .setFitFieldWidths(true)
     *      .setDefaultHeight(5)
     *      .setColumnValueOptions([column1: ["option1", "option2"]])
     *      .setCanModifyRows(false)
     *      .setNoCellRefresh(true)
     *      .setFixTableHeight(true)
     *      .buildContextParameter()
     * </pre></blockquote>
     *
     * @param value true to enable, false to disable rows modification
     * @return self
     */
    public InputMatrixInputBuilder setCanModifyRows(boolean value) {
        cp.setConfigParameter("canModifyRows", value);
        return this;
    }

    /**
     * Set cell refresher flag of the configuration parameters.
     * <p>Example:
     * <blockquote><pre>
     * api.inputBuilderFactory()
     *      .createInputMatrix("input")
     *      .setColumns(["column1", "column2"])
     *      .setReadOnlyColumns(["readOnly1", "readOnly2"])
     *      .setFitFieldWidths(true)
     *      .setDefaultHeight(5)
     *      .setColumnValueOptions([column1: ["option1", "option2"]])
     *      .setCanModifyRows(false)
     *      .setNoCellRefresh(true)
     *      .setFixTableHeight(true)
     *      .buildContextParameter()
     * </pre></blockquote>
     *
     * @param value true to disable, false to enable cell refresher
     * @return self
     */
    public InputMatrixInputBuilder setNoCellRefresh(boolean value) {
        cp.setConfigParameter("noCellRefresh", value);
        return this;
    }

    /**
     * Set fix table height flag of the configuration parameters.
     * <p>Example:
     * <blockquote><pre>
     * api.inputBuilderFactory()
     *      .createInputMatrix("input")
     *      .setColumns(["column1", "column2"])
     *      .setReadOnlyColumns(["readOnly1", "readOnly2"])
     *      .setFitFieldWidths(true)
     *      .setDefaultHeight(5)
     *      .setColumnValueOptions([column1: ["option1", "option2"]])
     *      .setCanModifyRows(false)
     *      .setNoCellRefresh(true)
     *      .setFixTableHeight(true)
     *      .buildContextParameter()
     * </pre></blockquote>
     *
     * @param value true to enable, false to disable fix table height
     * @return self
     */
    public InputMatrixInputBuilder setFixTableHeight(boolean value) {
        cp.setConfigParameter("fixTableHeight", value);
        return this;
    }

}
