package net.pricefx.lmformulaengine;

import java.util.*;
import java.util.Map.Entry;

import net.pricefx.common.api.PriceFxInterface;
import net.pricefx.formulaengine.ResultOverride;
import net.pricefx.lmformulaengine.model.MapExtended;
import net.pricefx.lmformulaengine.scripting.LmSandboxAPI;
import net.pricefx.logic.core.model.FormulaElementResult;
import net.pricefx.logic.core.model.FormulaTraceMessage;
import net.pricefx.server.dto.calculation.ContextParameter;

public class LmFormulaEngineContext {

    public final static int LM_LIMIT_ROWS = 100;
    public final static String CHECK_RESULT_STRING = "0.18057742339";
    public final static String CHECK_RESULT_TIME = "00:00";
    public final static String CHECK_RESULT_DATE = "2011-01-11";
    public final static String CHECK_RESULT_DATE_TIME = "2011-01-11T00:00:00";
    public final static String[] CHECK_RESULT_DATERANGE = {"2011-01-11","2011-01-11"};

    private boolean abortCalculation = false;
    private boolean keepOldOnAbortCalculation;
    private boolean abortSyntaxCheck;

    private String formulaNature;

    private String secondaryKey;
    private Map<String, Object> parameterMap = new HashMap();
    private Map<String, ResultOverride> manualResultOverrides;
    private Date targetDate;
    private String formulaElementGroup;
    private List<ContextParameter> missingParameters = new ArrayList();
    private List<String> warnings;

    private List<FormulaTraceMessage> traceMessages;
    private boolean traceEnabled = true;

    private String originalProductSKU; //holds the SKU the context was initialized with
    private String tempProductSKU; //holds a temporary sku context. Engine makes sure that this does not pass element boundaries
    private MapExtended currentItem;

    private Object finalResult;
    private Object intermediateResult;

    private String errorElementName;
    private String errorMessage;
    private int errorStartPosition;
    private int errorEndPosition;

    private Map<String,FormulaElementResult> resultElements = new HashMap();
    private List<FormulaElementResult> orderedResults = new ArrayList();
    private Map<String,Object> resultsByElementLabel = new HashMap();

    private String simulationSet;
    private boolean simulationMode;
    private Set<Long> simulationLookups;

    private boolean allowObjectModifications;

    private List<String[]> batchInfo;

    // we hold the random generator on ctx level so as to not destroy the randomness from one formula/formula element evaluation
    // to another.
    private Random random;

    //the following two are used to trigger alerts by groovy code
    private String alertMessage;
    private String alertType;

    /**
     * These will be made available in the engine for groovy scripts
     */
    private Map<String,Object> groovyBindings;
    private Map<String,Object> groovyGlobals;
    private LmSandboxAPI sandboxAPI;

    private boolean syntaxCheckMode = false;
    private boolean debugMode = false;
    private boolean distributedMode = false;
    private boolean currentItemDirty = false;

    // timeout as specified in formula element.
    private Integer elementTimeout;

    private Map<String, Object> product = null;
    private Map<String, Object> customer = null;

    private int lmLimitRows;

    /**
     * Element results the formula has access to.
     * UseCase: regular run results of formula are presented to invariant formula
     * via regular sandboxApi, better this way then bindings
     */
    private List<FormulaElementResult> formulaEngineResults = new ArrayList<>();

    public LmFormulaEngineContext(Date targetDate, String productSKU, String formulaNature) {
        this.targetDate = targetDate;
        this.originalProductSKU = productSKU;
        this.formulaNature = formulaNature;
        this.finalResult = Integer.valueOf(0);
        lmLimitRows = LM_LIMIT_ROWS;
    }

    public String getFormulaNature() {
        return formulaNature;
    }

    public void setFormulaNature(String formulaNature) {
        this.formulaNature = formulaNature;
    }

    public String getSecondaryKey() {
        return secondaryKey;
    }

    public void setSecondaryKey(String secondaryKey) {
        this.secondaryKey = secondaryKey;
    }

    public List<String[]> getBatchInfo() {
        return batchInfo;
    }

    public void setBatchInfo(List<String[]> batchInfo) {
        this.batchInfo = batchInfo;
    }

    public String getFormulaElementGroup() {
        return formulaElementGroup;
    }


    public void setFormulaElementGroup(String formulaElementGroup) {
        this.formulaElementGroup = formulaElementGroup;
    }

    public void resetAll(Date targetDate, String productSKU) {
        if(groovyBindings != null)
            groovyBindings.clear();

        resetAllWithoutLookupCache(targetDate, productSKU);

        elementTimeout = null;
        batchInfo = null;
    }

    public void resetAllWithoutLookupCache(Date targetDate, String productSKU) {
        parameterMap.clear();
        simulationSet = null;
        simulationMode = false;
        allowObjectModifications = false;
        simulationLookups = null;

        resetWithoutParameters(targetDate, productSKU);
    }

    public void resetWithoutParameters (Date targetDate, String productSKU) {
        resetResults();
        if(manualResultOverrides != null)
            manualResultOverrides.clear();

        this.targetDate = targetDate;
        this.originalProductSKU = productSKU;

        currentItem = null;
    }

    public void resetResults () {
        missingParameters.clear();
        resultElements.clear();
        orderedResults.clear();
        resultsByElementLabel.clear();

        resetSKUContextSwitch();
        this.finalResult = Integer.valueOf(0);
        clearWarnings();
        clearTraceMessages();

        finalResult = null;
        intermediateResult = null;
        errorMessage = null;
        errorStartPosition = 0;
        errorEndPosition = 0;
        sandboxAPI = null;
        abortCalculation = false;
        keepOldOnAbortCalculation = false;
        abortSyntaxCheck = false;
        currentItemDirty = false;
    }

    public void addResult(FormulaElementResult r) {
        if(r != null) {
            if(! resultElements.containsKey(r.getElementName())) {	//prevent overwriting
                resultElements.put(r.getElementName(), r);
                orderedResults.add(r);

                if (r.getElementLabel() != null){
                    resultsByElementLabel.put(r.getElementLabel(), r.getElementResult());
                }
            }
        }
    }

    public Random getRandom(){
        if (random == null){
            random = new Random();

        }
        return random;
    }

    public boolean isDebugMode() {
        return debugMode;
    }

    public void setDebugMode(boolean debugMode) {
        this.debugMode = debugMode;
    }

    public boolean isDistributedMode() {
        return distributedMode;
    }

    public void setDistributedMode(boolean distributedMode) {
        this.distributedMode = distributedMode;
    }

    /**
     * Gets a previous intermediate result value. Null if it cannot be found
     *
     * @param name
     * @return Value or null
     */
    public Object getResultValue(String name) {
        FormulaElementResult r = resultElements.get(name);
        if(r == null)
            return null;
        else
            return r.getElementResult();
    }

    public boolean resultExists(String name) {
        return resultElements.containsKey(name);
    }

    public Object getResultValueByElementLabel(String label) {
        if (label != null){
            return resultsByElementLabel.get(label);
        }
        return null;
    }

    public Map<String, FormulaElementResult> getResultElements() {
        return resultElements;
    }

    public Map<String, Object> getResultElementsResultMap() {
        Map<String,Object> m = new HashMap<String,Object>();
        for(String key : resultElements.keySet()) {
            m.put(key,getResultValue(key));
        }
        return m;
    }

    public List<FormulaElementResult> getOrderedResultElements() {
        return orderedResults;
    }

    public List<FormulaElementResult> getExternalOrderedResultElements() {
        if(orderedResults == null) return null;
        List<FormulaElementResult> externalResuls = new ArrayList<FormulaElementResult>();
        for(FormulaElementResult fer : orderedResults) {
            externalResuls.add(fer);
        }
        return externalResuls;
    }

    public FormulaElementResult getResult(String name) {
        return resultElements.get(name);
    }

    public void setSKUContext(String sku) {
        originalProductSKU = sku;
    }

    public String getProductSKU() {
        return tempProductSKU != null ? tempProductSKU : originalProductSKU;
    }

    public String getOriginalProductSKU() {
        return originalProductSKU;
    }

    public void switchSKUContext(String tempSKU) {
        tempProductSKU = tempSKU;
    }

    public void resetSKUContextSwitch() {
        tempProductSKU = null;
    }

    public Object getIntermediateResult() {
        return intermediateResult;
    }

    public void setIntermediateResult(Object intermediateResult) {
        this.intermediateResult = intermediateResult;
    }

    public Object getFinalResult() {
        return finalResult;
    }

    public void setFinalResult(Object finalResult) {
        this.finalResult = finalResult;
    }

    public Object getManualResultOverrideValue(String elementName) {
        if(manualResultOverrides == null)
            return null;
        if(!manualResultOverrides.containsKey(elementName))
            return null;

        return manualResultOverrides.get(elementName).getValue();
    }

    public void removeManualResultOverride(String elementName) {
        if(manualResultOverrides == null) return;

        if(!manualResultOverrides.containsKey(elementName)) return;

        manualResultOverrides.remove(elementName);
    }

    public boolean isForcedManualResultOverride(String elementName) {
        if(manualResultOverrides == null)
            return false;
        if(!manualResultOverrides.containsKey(elementName))
            return false;

        return manualResultOverrides.get(elementName).isForced();
    }

    public void addManualResultOverride(String elementName, Object value, boolean force) {
        if(manualResultOverrides == null)
            manualResultOverrides = new HashMap<String,ResultOverride>();

        manualResultOverrides.put(elementName, new ResultOverride(value,force));
    }

    public void addGroovyBinding(String bindingName, Object value) {
        if(groovyBindings == null)
            groovyBindings = new HashMap<String,Object>();

        if("api".equals(bindingName))
            return;

        groovyBindings.put(bindingName, value);
    }

    public Map<String, Object> getGroovyGlobals() {
        return groovyGlobals;
    }

    public void setGroovyGlobals(Map<String, Object> groovyGlobals) {
        this.groovyGlobals = groovyGlobals;
    }

    public LmSandboxAPI getSandboxAPI() {
        return sandboxAPI;
    }

    public void setSandboxAPI(LmSandboxAPI sandboxAPI) {
        this.sandboxAPI = sandboxAPI;
    }

    public Map<String, Object> getGroovyBindings() {
        return groovyBindings;
    }

    public void addManualResultOverride(ResultOverride ro, boolean force) {
        if(ro != null && ro.getElementName() != null) {
            addManualResultOverride(ro.getElementName(), ro.getValue(), force);
        }
    }

    public void abortCalculation(boolean keepOld) {
        abortCalculation = true;
        keepOldOnAbortCalculation = keepOld;
    }

    public boolean isAborted() {
        return abortCalculation;
    }

    public boolean isKeepOldOnAbortCalculation() {
        return keepOldOnAbortCalculation;
    }

    public void abortSyntaxCheck() {
        abortSyntaxCheck = true;
    }

    public boolean isAbortSyntaxCheck() {
        return abortSyntaxCheck;
    }

    /**
     * Retrieves a parameter from the context. If none found returns null
     * If a non-null value is returned, this parameter is automatically deemed to be used and put into
     * the user parameters collection
     * @param key
     * @return
     */
    public Object getParameter(String key){
        Object o = parameterMap.get(key);
        return o;
    }

    public Collection<String> getParameterNames(){
        return parameterMap.keySet();
    }

    /**
     * This returns all parameters, however in a cloned map to avoid that stuff is added/removed
     * Obviously the parameter value could be changed, but that was possible so far as well
     * I.e. no complete immutability
     * @return
     */
    public Map<String, Object> getParametersClone() {
        return new HashMap<String, Object>(parameterMap);

    }
    public void addParameter(String key, Object value) {
        if(PriceFxInterface.CONTEXTPARAM_NAME_PRODUCTSKU.equals(key) && originalProductSKU == null && value != null)
            originalProductSKU = value.toString();

        parameterMap.put(key, value);
    }

    public void addParameters(Map <String, Object> parameters) {
        if(parameters != null) {
            for (Entry<String, Object> parameter : parameters.entrySet()) {
                addParameter(parameter.getKey(), parameter.getValue());
            }
        }
    }

    public Date getTargetDate() {
        return targetDate;
    }

    public List<ContextParameter> getMissingParameters() {
        return missingParameters;
    }

    public void addMissingParameter(ContextParameter param) {
        if(missingParameters.contains(param))
            return;
        else
            missingParameters.add(param);
    }

    public ContextParameter getMissingParameter(String name) {
        if(missingParameters == null || name == null)
            return null;

        for(ContextParameter cp : missingParameters) {
            if(name.equals(cp.getName()))
                return cp;
        }
        return null;
    }

    public List<String> getWarnings() {
        return warnings;
    }

    public void clearWarnings() {
        if(warnings == null)
            return;

        warnings.clear();
    }

    public void addWarning(String param) {
        if(warnings == null)
            warnings = new ArrayList<String>();

        warnings.add(param);
    }

    public void addTraceMessage(FormulaTraceMessage msg) {
        if(!traceEnabled || msg == null)
            return;

        if(traceMessages == null)
            traceMessages = new ArrayList<FormulaTraceMessage>();

        traceMessages.add(msg);
    }

    public List<FormulaTraceMessage> getTraceMessages() {
        return traceMessages;
    }

    public void clearTraceMessages() {
        if(traceMessages == null)
            return;

        traceMessages.clear();
    }

    public boolean isTraceEnabled() {
        return traceEnabled;
    }

    public void setTraceEnabled(boolean traceEnabled) {
        this.traceEnabled = traceEnabled;
    }

    public void clearAlert() {
        alertMessage = null;
        alertType = null;
    }

    public boolean hasAlert() {
        return alertType != null
                || alertMessage != null;
    }

    public String getAlertMessage() {
        return alertMessage;
    }

    public void setAlertMessage(String alertMessage) {
        this.alertMessage = alertMessage;
    }

    public String getAlertType() {
        return alertType;
    }

    public void setAlertType(String alertType) {
        this.alertType = alertType;
    }

    public int getErrorStartPosition() {
        return errorStartPosition;
    }

    public void setErrorStartPosition(int errorStartPosition) {
        this.errorStartPosition = errorStartPosition;
    }

    public int getErrorEndPosition() {
        return errorEndPosition;
    }

    public void setErrorEndPosition(int errorEndPosition) {
        this.errorEndPosition = errorEndPosition;
    }

    public String getErrorMessage() {
        return errorMessage;
    }

    public void setErrorMessage(String errorMessage) {
        this.errorMessage = errorMessage;
    }

    public String getErrorElementName() {
        return errorElementName;
    }

    public void setErrorElementName(String errorElementName) {
        this.errorElementName = errorElementName;
    }

    public boolean isSimulation() {
        return simulationMode;
    }

    public void setSimulation(boolean isSim) {
        if(!isSim) this.simulationSet = null;
        this.simulationMode = isSim;
    }

    public String getSimulationSet() {
        return simulationSet;
    }

    public void addSimulationOnlyLookup(Long id) {
        if(simulationLookups == null)
            simulationLookups = new HashSet<Long>();

        simulationLookups.add(id);
    }

    public Set<Long> getSimulationOnlyLookups() {
        return simulationLookups;
    }

    public void setSimulationSet(String simulationSet) {
        if(simulationSet != null && simulationSet.isEmpty())
            this.simulationSet = null;
        else
            this.simulationSet = simulationSet;

        if(this.simulationSet != null) simulationMode = true;
    }

    public boolean isSyntaxCheckMode() {
        return syntaxCheckMode;
    }

    public MapExtended getCurrentItem() {
        return currentItem;
    }

    public void setCurrentItem(MapExtended currentItem) {
        this.currentItem = currentItem;
    }

    public void setCurrentItemDirty() {
        currentItemDirty = true;
    }

    public void setCurrentItemClean() {
        currentItemDirty = false;
    }

    public boolean isCurrentItemDirty() {
        return currentItemDirty;
    }

    public Integer getElementTimeout() {
        return elementTimeout;
    }

    public void setElementTimeout(Integer elementTimeout) {
        this.elementTimeout = elementTimeout;
    }

    public boolean allowObjectModifications() {
        return allowObjectModifications;
    }

    public void setAllowObjectModifications(boolean allowObjectModifications) {
        this.allowObjectModifications = allowObjectModifications;
    }

    public List<FormulaElementResult> getFormulaEngineResults() {
        return formulaEngineResults;
    }

    public void setFormulaEngineResults(List<FormulaElementResult> formulaEngineResults) {
        this.formulaEngineResults = formulaEngineResults;
    }

    public Map<String, Object> getProduct() {
        return product;
    }

    public void setProduct(Map<String, Object> product) {
        this.product = product;
    }

    public Map<String, Object> getCustomer() {
        return customer;
    }

    public void setCustomer(Map<String, Object> customer) {
        this.customer = customer;
    }

    public int getLmLimitRows() {
        return lmLimitRows;
    }

    public void setLmLimitRows(int lmLimitRows) {
        this.lmLimitRows = lmLimitRows;
    }

}