package net.pricefx.server.dto.calculation;

import java.io.Serializable;
import java.util.*;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.googlecode.genericdao.search.Filter;

import net.pricefx.common.api.InputType;
import net.pricefx.common.api.PriceFxInterface;
import net.pricefx.domain.LookupTable;
import net.pricefx.logic.core.Utils;
import net.pricefx.service.impl.AdvancedFilterUtil;

/**
 * This class represents the input parameter, the input field. So you can use it to set its label, default value, if required, etc.
 */
@JsonIgnoreProperties(ignoreUnknown = true) 	//need this to ignore client-only props that we don't need server side but are part of the request
public class ContextParameter implements Serializable {

	private static final long serialVersionUID = 3443538670067930182L;

	private String name;
	private String label;
	private String lookupTableId;
	private String url;
	private InputType type;
	private Object value;
	private String valueHint;
	private String[] valueOptions;
	private Boolean readOnly;
	private Map<String, Object> filter;	//e.g. to filter down a LT
	private String parameterGroup;
	private Boolean required;
	private String labelTranslations;
	private Boolean addUnknownValues;
	private String typedId; // context typed id, i.e. Quote typed id when quote pricing running, ...
	private Boolean alwaysEditable;
	private List<ContextParameter> inputs = new ArrayList<>();

	//extra config options for the parameter (typically as Map)
	//strictly other existing fields would also fall into this category but we omit that due to backward compatibility
	//Adds additional configuration parameter/option to the input. These options modifies, how the input field *behaves*.
	private Map<String, Object> parameterConfig;

	//Adds additional configuration options that influence how the inpoot *looks like*
	private Map<String,Object> formattingOptions;


	//need that for Quote deserialization with Jackson
	public ContextParameter() {}

	public ContextParameter(InputType type, String name, Date targetDate, Object objService) {
		this.name = name;
		this.type = type;

		// get some more info for the context parameter
		// the basic assumption is that the name is either
		// - a parameter table name
		// - the customer special name
		if(type == null)
			return;

		if(type.equals(InputType.CUSTOMER)){
			label = "CUSTOMER";
			url = "/" + PriceFxInterface.CMD_FETCH + "/C/"; 
		} else if(type.equals(InputType.CUSTOMERGROUP) && PriceFxInterface.CONTEXTPARAM_NAME_CUSTOMERGROUP.equals(name)){
			label = "CUSTOMERGROUP";
		} else if(type.equals(InputType.PRODUCTGROUP) && PriceFxInterface.CONTEXTPARAM_NAME_PRODUCTGROUP.equals(name)){
			label = "PRODUCTGROUP";
		} else if (type.equals(InputType.LOOKUP)){
			//here we need to do some lookups
			// NOTE: should use other constructor really
			LookupTable lt = null;
			if(lt != null){
				label = lt.getLabel();
				url = "/" + PriceFxInterface.CMD_LTM_FETCH + "/" + lt.getId();
			}
		} else {
			label = name;

			// automatically set the config dataType so that the user won't have to
			// - as it defaults to CONTEXTPARAM_CONFIG_DATATYPE_REAL on the client - if not set
			switch (type) {
				case USERENTRY:
					setConfigParameter(PriceFxInterface.FIELD_CONTEXTPARAM_CONFIG_DATATYPE,
							PriceFxInterface.CONTEXTPARAM_CONFIG_DATATYPE_REAL);
					break;
				case INTEGERUSERENTRY:
					setConfigParameter(PriceFxInterface.FIELD_CONTEXTPARAM_CONFIG_DATATYPE,
							PriceFxInterface.CONTEXTPARAM_CONFIG_DATATYPE_INTEGER);
					break;
				case STRINGUSERENTRY:
					setConfigParameter(PriceFxInterface.FIELD_CONTEXTPARAM_CONFIG_DATATYPE,
							PriceFxInterface.CONTEXTPARAM_CONFIG_DATATYPE_STRING);
					break;
				case DATETIMEUSERENTRY:
					setConfigParameter(PriceFxInterface.FIELD_CONTEXTPARAM_CONFIG_DATATYPE,
							PriceFxInterface.CONTEXTPARAM_CONFIG_DATATYPE_DATETIME);
					break;
				case DATEUSERENTRY:
					setConfigParameter(PriceFxInterface.FIELD_CONTEXTPARAM_CONFIG_DATATYPE,
							PriceFxInterface.CONTEXTPARAM_CONFIG_DATATYPE_DATE);
					break;
				case TIMEUSERENTRY:
					setConfigParameter(PriceFxInterface.FIELD_CONTEXTPARAM_CONFIG_DATATYPE,
							PriceFxInterface.CONTEXTPARAM_CONFIG_DATATYPE_TIME);
					break;
				case PRODUCT:
					setConfigParameter(PriceFxInterface.FIELD_CONTEXTPARAM_CONFIG_DATATYPE,
							PriceFxInterface.CONTEXTPARAM_CONFIG_DATATYPE_PRODUCT);
					break;
				case CUSTOMER:
					setConfigParameter(PriceFxInterface.FIELD_CONTEXTPARAM_CONFIG_DATATYPE,
							PriceFxInterface.CONTEXTPARAM_CONFIG_DATATYPE_CUSTOMER);
					break;
				default: // at the moment, only the types above can be customized via the config feature
					break; // => nothing to do here
			}
		}

	}

	public ContextParameter(InputType type, String name, Object value) {
		this.type = type;
		this.name = name;
		this. value = value;
	}

	public ContextParameter(LookupTable lt, Date targetDate) {
		type = lt.getNumberOfKeyFields() > 1 ? InputType.MATRIXLOOKUP : InputType.LOOKUP;
		name = lt.getUniqueName();
		updateWithLookupTable(lt);
	}

	public ContextParameter(LookupTable lt, Date targetDate, String defaultValue) {
		this(lt, targetDate);
		this.value = defaultValue;
	}

	public ContextParameter(String name, Object fc, String field, String defaultValue) {
		this.type = InputType.DMDIMFILTER;
		this.name = name;
		this.value = defaultValue;
//		this.url = fc.getTypedId() + "/" + field;
	}

	public void addParameterConfigEntry(String key, Object value) {
		setConfigParameter(key,value);
	}
	/**
	 * This is a convenience method to update a ContextParameter with a new version of the same lookup table
	 * This can happen e.g. if a quote is worked on a while and passes a validity date border
	 * or an old quote is copied
	 *
	 * @param lt
	 */
	public void updateWithLookupTable(LookupTable lt) {
		 if (!isLookupType())
			 return;
		 if(lt == null)
			 return;
		 if(name != null && !name.equals(lt.getUniqueName()))
			 return;

		 url = "/" + PriceFxInterface.CMD_LTM_FETCH + "/" + lt.getId();
		 lookupTableId = lt.getId().toString();
		 label = lt.getLabel();
	}

	@JsonIgnore
	public boolean isLookupType() {
		return InputType.LOOKUP.equals(type) || InputType.MATRIXLOOKUP.equals(type);
	}

	@JsonIgnore
	public final boolean isCustomerParam() {
		return InputType.CUSTOMER.equals(type);
	}

	@JsonIgnore
	public final boolean isProductParam() {
		return InputType.PRODUCT.equals(type);
	}

	@JsonIgnore
	public final boolean isProductGroupParam() {
		return InputType.PRODUCTGROUP.equals(type);
	}

	@JsonIgnore
	public final boolean isMatrixParam() {
		return InputType.MATRIXLOOKUP.equals(type);
	}

	@JsonIgnore
	public final boolean isOptionParam() {
		return InputType.OPTION.equals(type);
	}

	@JsonIgnore
	public final boolean isOptionsParam() {
		return InputType.OPTIONS.equals(type);
	}

	@JsonIgnore
	public final boolean isHiddenParam() {
		return InputType.HIDDEN.equals(type);
	}

	public Boolean isRequired() { return required;	}

	/**
	 * Sets, if it's mandatory for the input parameter to have a value. The value can be either entered by user or a default value.
	 * If the input value is required, but not provided, then it's usually somehow expressed on the screen (usually exclamation mark).
	 * <p>In case of Quote, all required fields must be populated, otherwise you cannot submit the Quote for approval.
	 * @param required If the input value will be required or not.
	 */
	public ContextParameter setRequired(Boolean required) {
		this.required = required;
		return this;
	}

	public String getName() {
		return name;
	}

	public String getLabel() {
		return label;
	}

	/**
	 * Sets the label of the input parameter to the given text.
	 * @param label The text to be displayed as label of the input field.
	 */
	public ContextParameter setLabel(String label) {
		this.label = label;
		return this;
	}

	public String getLabelTranslations() {
		return labelTranslations;
	}

	public ContextParameter setLabelTranslationsFromMap(Map<String,String> translations) throws JsonProcessingException {
		labelTranslations = Utils.getObjectMapper().writeValueAsString(translations);
		return this;
	}

	public ContextParameter setLabelTranslations(String labelTranslations) {
		this.labelTranslations = labelTranslations;
		return this;
	}

	public Map<String, Object> getFilter() {
		return filter;
	}

	public ContextParameter setFilter(Map<String, Object> filter) {
		this.filter = filter;
		return this;
	}

	@JsonIgnore
	public ContextParameter setFilters(Filter ... filters) {
		setFilter(AdvancedFilterUtil.toAdvancedCriterion(Filter.OP_AND, filters));
		return this;
	}

	public String getParameterGroup() {
		return parameterGroup;
	}

	/**
	* Allows you to create collapsible sections that organize inputs in the UI. You can assign parameterGroup
	* to each parameter and they will get categorized. The order of parameter groups (sections) is based on the order
	* in which they are "created" (declared) when the logic is executed. There is one default group for all those that
	* do not have parameterGroup assigned.
	*
	* <p> This is supported for Quotes, Contracts and Rebate Agreements (not for Configurators).
	*/
	public ContextParameter setParameterGroup(String parameterGroup) {
		this.parameterGroup = parameterGroup;
		return this;
	}

	public String getURL() {
		return url;
	}

	public ContextParameter setURL(String url) {
		this.url = url;
		return this;
	}

	public String getType() {
		return type == null ? "" : type.name();
	}

	@JsonIgnore
    public InputType getInputType() {
        return type;
    }

    public ContextParameter applyType(InputType inputType) {
		this.type = inputType;
		return this;
	}

	public String getValueHint() {
		return valueHint;
	}

	public ContextParameter setValueHint(String hint) {
		valueHint = hint;
		return this;
	}

	@JsonProperty(value="valueOptions")
	public String[] getValueOptions() {
		return valueOptions;
	}

	/**
	 * Sets list of options used in the dropdown input field. This works only for input parameters created via {@code api.option()} and {@code api.options()}
	 * <p>Example:
	 * <blockquote><pre>
	 * api.option("Country")
	 * // ...
	 * api.getParameter("Country").setValueOptions(["USA", "DE"])
	 * </pre></blockquote>
	 * @param valueOptions
	 */
	public ContextParameter setValueOptions(String[] valueOptions) {
		this.valueOptions = valueOptions;
		return this;
	}

	@JsonIgnore
	public ContextParameter setValueOptions(List<String> valueOptions) {
		this.valueOptions = valueOptions != null ? valueOptions.toArray(new String[]{}) : null;
		return this;
	}

	public String getLookupTableId(){
		return lookupTableId;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		ContextParameter other = (ContextParameter) obj;
		//and same name
		if (name == null) {
			if (other.getName() != null)
				return false;
		} else if (!name.equals(other.getName()))
			return false;

		return true;
	}

	public Object getValue() {
		return value;
	}

	/**
	 * Set initial/default value of the input parameter.
	 * <p><strong>Important:</strong> this makes sense only during the SyntaxCheck execution mode.
	 *                      You cannot set the initial value of input field later during regular execution.</p>
	 * @param value
	 */
	public ContextParameter setValue(Object value) {
		this.value = value;
		return this;
	}

	public Boolean getReadOnly() {
		return readOnly;
	}

	public ContextParameter setReadOnly(Boolean readOnly) {
		this.readOnly = readOnly;
		return this;
	}

	public ContextParameter setParameterConfig(Map<String, Object> parameterConfig) {
		this.parameterConfig = parameterConfig;
		return this;
	}

	public Map<String, Object> getParameterConfig() {
		if(parameterConfig == null) {
			parameterConfig = new HashMap<>();
		}
		return parameterConfig;
	}


	/**
	 * Adds additional configuration parameter/option to the input. These options modifies, how the input field behaves.
	 * <p>You can use following {@code paramName} and their values
	 * <ul>
	 *     <li><strong>inputType</strong> - e.g. "range", ...</li>
	 *     <li><strong>formatType</strong> - e.g. "PERCENT" (the input will be formatted with % sign and the value returned will be 0..1 instead of 0..100), ...</li>
	 *     <li><strong>from</strong> - the minimum value of the allowed value range. Works only for inputType "range".</li>
	 *     <li><strong>to</strong> - the maximum value of the allowed value range. Works only for inputType "range".</li>
	 *     <li><strong>labels</strong> - labels to be used for the options, e.g. ["USA":"United States of America, "DE":"Germany]</li>
	 * </ul>
	 * @param paramName
	 * @param value
	 * @return
	 *
	 * @see #addParameterConfigEntry(String, Object)
	 */
	public ContextParameter setConfigParameter(String paramName, Object value) {
		getParameterConfig().put(paramName, value);
		return this;
	}

	public void clearConfigParameter(String paramName) {
		getParameterConfig().put(paramName, null);
	}

	public Map<String, Object> getFormattingOptions() {
		if(formattingOptions == null) {
			formattingOptions = new HashMap<>();
		}
		return formattingOptions;
	}

	public void setFormattingOptions(Map<String, Object> formattingOptions) {
		this.formattingOptions = formattingOptions;
	}

	public ContextParameter setFormattingOption(String optionName, Object value) {
		getFormattingOptions().put(optionName, value);
		return this;
	}

	public Boolean getAddUnknownValues() {
		return addUnknownValues;
	}

	public ContextParameter setAddUnknownValues(Boolean addUnknownValues) {
		this.addUnknownValues = addUnknownValues;
		return this;
	}

	public String getTypedId() {
		return typedId;
	}

	public ContextParameter setTypedId(String typedId) {
		this.typedId = typedId;
		return this;
	}

	public Boolean getAlwaysEditable() {
		return alwaysEditable;
	}

	public ContextParameter setAlwaysEditable(Boolean alwaysEditable) {
		this.alwaysEditable = alwaysEditable;
		return this;
	}

	/**
	 * Adds a nested input. This input can be of any type. This is intended to be used when the parent input is
	 * of type @link InputType#ROW} or type {@link InputType#COLLAPSE}. Depending on the parent type, it is also
	 * expected that the added input will have a flex attribute.
	 * @param contextParameter the input to add
	 */
	public void addInput(ContextParameter contextParameter) {
		this.inputs.add(contextParameter);
	}

	public List<ContextParameter> getInputs() {
		return inputs;
	}

	@Override
	public int hashCode() {
	    int hash = 1;
	    hash = hash * 31 + (name == null ? 0 : name.hashCode());
	    return hash;
	}

	@Override
	public String toString() {
		return (name == null ? "null" : name) + " (" + getType() + ")";
	}


}
