package net.pricefx.server.dto.calculation;

import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.fasterxml.jackson.annotation.JsonProperty;
import net.pricefx.common.api.FieldFormatType;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

@JsonIgnoreProperties(ignoreUnknown = true)
public class ResultMatrixDto extends AbstractPortletDto implements ResultMatrix {

    private static final long serialVersionUID = 900L;

    private final List<Map<String, Object>> entries;
    private final List<String> columns;
    private FieldFormatType defaultFormat;
    private Map<String, FieldFormatType> columnFormats;
    private Map<String, String> columnTooltips;
    private boolean enableClientFilter;
    private String title;
    private String preferenceName;
    private boolean preferencesDisabled;
    private boolean sortingDisabled;
    private boolean escapeHTML;
    private Map<String, String> columnTranslations;

    @JsonProperty
    private final Map<String, Map<String, String>> onRowSelectEvents = new HashMap<>();

    @JsonProperty
    private List<ResultMatrix.BackEndAction> backEndActions =   new ArrayList<>();

    public ResultMatrixDto() {
        this((Map<?, ?>) null);
    }

    @SuppressWarnings({"rawtypes", "unchecked"})
    public ResultMatrixDto(Map jsonMap) {
        if (jsonMap == null) {
            entries = new ArrayList<>();
            columns = new ArrayList<>();
        } else {
            entries = (List<Map<String, Object>>) jsonMap.get("entries");
            columns = (List<String>) jsonMap.get("columns");
        }

    }

    @Override
    public String getTitle() {
        return title;
    }

    @Override
    public ResultMatrix withTitle(String title) {
        this.title = title;
        return this;
    }

    @Override
    public String getPreferenceName() {
        return preferenceName;
    }

    @Override
    public boolean isPreferencesDisabled() {
        return preferencesDisabled;
    }


    @Override
    public ResultMatrix withDisablePreferences(boolean disablePreferences) {
        this.preferencesDisabled = disablePreferences;
        return this;
    }

    @Override
    public boolean isSortingDisabled() {
        return sortingDisabled;
    }

    @Override
    public ResultMatrix withDisableSorting(boolean disableSorting) {
        this.sortingDisabled = disableSorting;
        return this;
    }

    @Override
    public boolean isEscapeHTML() {
        return escapeHTML;
    }

    @Override
    public ResultMatrix withEscapeHTML(boolean escapeHTML) {
        this.escapeHTML = escapeHTML;
        return this;
    }

    @Override
    public ResultMatrix withPreferenceName(String preferenceName) {
        if (preferenceName != null)
            this.preferenceName = preferenceName.replaceAll("[^a-zA-Z0-9]+", "");
        else
            this.preferenceName = null;
        return this;
    }

    public ResultMatrixDto(String... columns) {
        this();
        for (int i = 0; i < columns.length; i++)
            addColumn(columns[i]);
    }

    @Override
    public ResultMatrix withColumnTranslation(String column, Map<String, String> translations) {
        if (columnTranslations == null) {
            columnTranslations = new HashMap<>();
        }

//        columnTranslations.put(column, DomainObject.jsonSerialize(translations));
        return this;

    }

    @Override
    public Map<String, String> getColumnTranslations() {
        return columnTranslations;
    }


    public ResultMatrixDto(Collection<String> columns) {
        this(columns.toArray(new String[columns.size()]));
    }

    @Override
    public ResultMatrix withColumn(String column) {
        if (!columns.contains(column))
            columns.add(column);
        return this;
    }

    @Override
    public ResultMatrix withColumn(String column, String tooltip) {
        addColumn(column);
        setColumnTooltip(column, tooltip);
        return this;
    }

    @Override
    public ResultMatrix withColumns(Collection<String> columns) {
        for (String col : columns) {
            addColumn(col);
        }
        return this;
    }

    @Override
    public void addRow(Map<String, Object> row) {
        entries.add(row);
    }

    @Override
    public ResultMatrix withRow(List<Object> row) {
        if (row != null) {
            Map<String, Object> m = new HashMap<>();
            for (int c = 0; c < Math.min(row.size(), columns.size()); c++) {
                m.put(columns.get(c), row.get(c));
            }
            entries.add(m);
        }
        return this;
    }

    @Override
    public ResultMatrix withRows(List<Map<String, Object>> rows) {
        if (rows != null) {
            for (Map<String, Object> row : rows) {
                addRow(row);
            }
        }
        return this;
    }

    @Override
    @SuppressWarnings("unchecked")
    public void addRow(Object... values) {
        if (values != null) {
            if (values.length == 1 && values[0] instanceof Map) {
                addRow((Map<String, Object>) values[0]);
            } else {
                addRow(Arrays.asList(values));
            }
        }
    }

    @Override
    public List<String> getColumns() {
        return columns;
    }

    @Override
    public List<Map<String, Object>> getEntries() {
        return entries;
    }

    @Override
    public Map<String, Object> findRow(String key, Object value) {
        for (Map<String, Object> entry : entries) {
            Object val = entry.get(key);
            if (val == null && value == null) return entry;
            if (val != null && val.equals(value)) return entry;
        }
        return null;
    }

    @Override
    public ResultMatrix withDefaultFormat(FieldFormatType fft) {
        defaultFormat = fft;
        return this;
    }

    @Override
    public FieldFormatType getDefaultFormat() {
        return defaultFormat;
    }

    @Override
    public Map<String, FieldFormatType> getColumnFormats() {
        return columnFormats;
    }

    @Override
    public ResultMatrix withColumnFormat(String column, FieldFormatType fft) {
        if (columnFormats == null)
            columnFormats = new HashMap<>();
        columnFormats.put(column, fft);
        return this;
    }

    @Override
    public Map<String, String> getColumnTooltips() {
        return columnTooltips;
    }

    @Override
    public ResultMatrix withColumnTooltip(String column, String tooltip) {
        if (columnTooltips == null)
            columnTooltips = new HashMap<>();

        if (tooltip == null)
            columnTooltips.remove(column);
        else
            columnTooltips.put(column, tooltip);
        return this;
    }

    @Override
    public boolean isEnableClientFilter() {
        return enableClientFilter;
    }

    @Override
    public ResultMatrix withEnableClientFilter(boolean enableClientFilter) {
        this.enableClientFilter = enableClientFilter;
        return this;
    }

    @Override
    public StyledCellDto styledCell(Object value) {
        StyledCellDto rmcc = new StyledCellDto();
        rmcc.value = value;
        return rmcc;
    }

    @Override
    public StyledCellDto styledCell(Object value, String textColor, String bgColor) {
        StyledCellDto rmcc = new StyledCellDto();
        rmcc.value = value;
        rmcc.color = textColor;
        rmcc.background = bgColor;
        return rmcc;
    }

    @Override
    public StyledCellDto styledCell(Object value, String textColor, String bgColor, String weight) {
        StyledCellDto rmsc = new StyledCellDto();
        rmsc.value = value;
        rmsc.color = textColor;
        rmsc.background = bgColor;
        rmsc.fontweight = weight;
        return rmsc;
    }

    @Override
    public StyledCellDto styledCell(Object value, String textColor, String bgColor, String weight, String alignment) {
        StyledCellDto rmsc = new StyledCellDto();
        rmsc.value = value;
        rmsc.color = textColor;
        rmsc.background = bgColor;
        rmsc.fontweight = weight;
        rmsc.alignment = alignment;
        return rmsc;
    }

    @Override
    public ImageCellDto imageCell(Object value) {
        ImageCellDto rmic = new ImageCellDto();
        rmic.value = value;
        return rmic;
    }

    @Override
    public ImageCellDto libraryImage(String mainName, String statusName) {
        ImageCellDto rmic = new ImageCellDto();
        rmic.value = "images/library/" + mainName + "_" + statusName + ".png";
        return rmic;
    }

    @Override
    public LinkCellDto linkCell(final Object value, final String targetPage, final Object targetPageState) {
        return new LinkCellDto(value, targetPage, targetPageState);
    }

    @Override
    public LinkButtonCellDto linkButtonCell(final Object value, final String targetPage,
                                                                     final Object targetPageState, final Object buttonState) {
        return new LinkButtonCellDto(value, targetPage, targetPageState, buttonState);
    }

    @Override
    public BackEndCellDto backEndAction(String label, String url, String payload, String successMessage, String failMessage) {
        BackEndCellDto cell = new BackEndCellDto();
        cell.label = label;
        cell.values = new String[]{url, payload, successMessage, failMessage};
        return cell;
    }

    @Override
    public BackEndActionDto rowSelectionBackEndAction(final String id) {
        final BackEndActionDto action = new BackEndActionDto(id);
        backEndActions.add(action);
        return action;
    }

    @Override
    public DownloadCellDto downloadButton(String label, String url, String payload) {
        DownloadCellDto cell = new DownloadCellDto();
        cell.label = label;
        cell.values = new String[]{url, payload};
        return cell;
    }

    @Override
    public CellsDto cells(String label, Object... cells) {
        CellsDto cs = new CellsDto();
        cs.label = label;
        cs.cells = cells;
        return cs;
    }

    @Override
    public OnRowSelectionDto onRowSelection() {
        return new OnRowSelectionDto();
    }

    @Override
    public String toString() {
        return String.format("(%s),(%s)", columns, entries);
    }

    protected FieldFormatType decideFormat(Object input) {
        if (input instanceof org.joda.time.LocalDate ||
                input instanceof java.time.LocalDate) {
            return FieldFormatType.DATE;
        }
        if (input instanceof org.joda.time.DateTime ||
                input instanceof java.time.LocalDateTime ||
                input instanceof org.joda.time.LocalDateTime ||
                input instanceof java.util.Calendar ||
                input instanceof java.util.Date ||
                input instanceof java.time.Instant ||
                input instanceof org.joda.time.Instant) {
            return FieldFormatType.DATETIME;
        }

        if (input instanceof Integer) {
            return FieldFormatType.INTEGER;
        }
        if (input instanceof Long) {
            return FieldFormatType.NUMERIC_LONG;
        }

        if (input instanceof Number) {
            return FieldFormatType.NUMERIC;
        }

        if (input instanceof String) {
            return FieldFormatType.TEXT;
        }

        return null;
    }

    @Override
    public void addFormatsToResultMatrix() {
        /*
        getEntries().stream() // one row is map of columns
                .flatMap(e -> e.entrySet().stream()) // stream of colums of all rows
                .filter(entry -> entry.getValue() != null)
                .map(entry -> new Pair<>(entry.getKey(), decideFormat(entry.getValue())))
                .filter(p -> p.getValue() != null)
                .forEach(p -> setColumnFormat(p.getKey(), p.getValue()));
            */
    }

    @Override
    public void normalizeDateTimes() {
        Set<String> datetimeKeys = columnFormats.entrySet().stream()
                .filter(e -> e.getValue() == FieldFormatType.DATETIME)
                .map(Map.Entry::getKey)
                .collect(Collectors.toSet());
        for (Map<String, Object> row : getEntries()) {
            for (Map.Entry<String, Object> column : row.entrySet()) {
                if (datetimeKeys.contains(column.getKey()) && column.getValue() != null) {
//                    column.setValue(DateTimeUtil.convert(column.getValue()));
                }
            }
        }

    }


    public static abstract class AbstractCellDto implements net.pricefx.server.dto.calculation.ResultMatrix.ResultMatrixCell {
        public Object value;
        public String type;
    }

    public static class StyledCellDto extends AbstractCellDto implements net.pricefx.server.dto.calculation.ResultMatrix.ResultMatrixStyledCell {
        public String color;
        public String background;
        public String fontweight;
        public String alignment;
        public String extraCSS;
        public String tooltip;

        public StyledCellDto() {
            type = "StyledCell";
        }

        @Override
        public StyledCell withColor(String color) {
            this.color = color;
            return this;
        }

        @Override
        public StyledCell withBackground(String bgColor) {
            this.background = bgColor;
            return this;
        }

        @Override
        public StyledCell withFontWeight(String fontWeight) {
            this.fontweight = fontWeight;
            return this;
        }

        @Override
        public StyledCell withAlignment(String alignment) {
            this.alignment = alignment;
            return this;
        }

        @Override
        public StyledCell withToolTip(String tooltip) {
            this.tooltip = tooltip;
            return this;
        }

        @Override
        public StyledCell withCSS(String extraCSS) {
            this.extraCSS = extraCSS;
            return this;
        }
    }

    public static class LinkCellDto extends AbstractCellDto implements net.pricefx.server.dto.calculation.ResultMatrix.ResultMatrixLinkCell {
        public String targetPage;
        public Object targetPageState;

        public LinkCellDto(final Object value, final String targetPage, final Object targetPageState) {
            type = "LinkCell";
            this.value = value;
            this.targetPage = targetPage;
            this.targetPageState = targetPageState;
        }

    }

    public static class LinkButtonCellDto extends LinkCellDto implements net.pricefx.server.dto.calculation.ResultMatrix.ResultMatrixLinkButtonCell {
        public Object buttonState; // this will soon allow to request the FE to display an optional dropDown

        public LinkButtonCellDto(final Object value, final String targetPage, final Object targetPageState,
                                          final Object buttonState) {
            super(value, targetPage, targetPageState);
            type = "LinkButtonCell";
            this.buttonState = buttonState;
        }
    }

    public static class BackEndCellDto implements net.pricefx.server.dto.calculation.ResultMatrix.ResultMatrixBackEndCell {
        public String type;
        public String label;
        public String[] values;

        public BackEndCellDto() {
            type = "BackEndAction";
        }
    }

    public static class DownloadCellDto implements net.pricefx.server.dto.calculation.ResultMatrix.ResultMatrixDownloadCell {
        public String type;
        public String label;
        public String[] values;

        public DownloadCellDto() {
            type = "DownloadButton";
        }
    }

    public static class CellsDto implements net.pricefx.server.dto.calculation.ResultMatrix.ResultMatrixCells {
        public Object[] cells;
        public String label;
        public String type;

        public CellsDto() {
            type = "Cells";
        }
    }

    public static class ImageCellDto extends AbstractCellDto implements net.pricefx.server.dto.calculation.ResultMatrix.ResultMatrixImageCell {

        public ImageCellDto() {
            type = "ImageCell";
        }
    }

    public class OnRowSelectionDto implements net.pricefx.server.dto.calculation.ResultMatrix.ResultMatrixOnRowSelection {

        public OnRowSelectionDto() {
        }

        @Override
        public TriggerEventDto triggerEvent(String eventName) {
            if (eventName == null || eventName.isEmpty()) {
//                throw new IllegalArgumentException(S.i18n(Msg.FE_ILLEGAL_ARGUMENT, "triggerEvent(...)"));
            }
            HashMap<String, String> mappings = new HashMap<>();
            onRowSelectEvents.put(eventName, mappings);
            return new TriggerEventDto(mappings);
        }

        public class TriggerEventDto implements net.pricefx.server.dto.calculation.ResultMatrix.ResultMatrixOnRowSelection.ResultMatrixTriggerEvent {

            private final Map<String, String> mappings;

            public TriggerEventDto(HashMap<String, String> mappings) {
                this.mappings = mappings;
            }

            @Override
            public TriggerEventDto withColValueAsEventDataAttr(String colName, String attrName) {
                if (colName == null || colName.isEmpty()) {
//                    throw new IllegalArgumentException(S.i18n(Msg.FE_ILLEGAL_ARGUMENT, "withColValueAsEventDataAttr(...)"));
                }
                if (attrName == null || attrName.isEmpty()) {
//                    throw new IllegalArgumentException(S.i18n(Msg.FE_ILLEGAL_ARGUMENT, "withColValueAsEventDataAttr(...)"));
                }
                mappings.put(colName, attrName);
                return TriggerEventDto.this;
            }
        }
    }

    public static class BackEndActionDto implements net.pricefx.server.dto.calculation.ResultMatrix.ResultMatrixBackEndAction {

        // we may want to add more types, but so far most useful one that handles it all is formulamanager.executeformula
        @JsonProperty
        private final String type = "formula";

        @JsonProperty
        private final String id;

        @JsonProperty
        private final LogicParameters parameters = new LogicParameters();

        private static class LogicParameters implements Serializable {
            @JsonProperty
            private String logicName;
            @JsonProperty
            private Date targetDate;
            @JsonProperty
            private final List<String> columns = new ArrayList<>();
            @JsonProperty
            private final Map<String, Object> additionalInputs = new HashMap<>();
        }

        // Visible elements in the UI / user feedback
        @JsonProperty
        private String buttonLabel;
        @JsonProperty
        private String buttonTooltip;
        @JsonProperty
        private String successMessage;
        @JsonProperty
        private String failureMessage;

        private BackEndActionDto() {
            this(null);
        }

        private BackEndActionDto(final String actionId) {
            this.id = actionId;
        }

        @Override
        public BackEndActionDto withLogicName(final String name) {
            this.parameters.logicName = name;
            return this;
        }

        @Override
        public BackEndActionDto withTargetDate(final Date date) {
            this.parameters.targetDate = date;
            return this;
        }

        @Override
        public BackEndActionDto withColumns(final String... names) {
            this.parameters.columns.clear();
            this.parameters.columns.addAll(Arrays.asList(names));
            return this;
        }

        @Override
        public BackEndActionDto addColumn(final String name) {
            this.parameters.columns.add(name);
            return this;
        }

        @Override
        public BackEndActionDto addFormulaInput(final String name, final Object value) {
            this.parameters.additionalInputs.put(name, value);
            return this;
        }

        @Override
        public BackEndActionDto withButtonLabel(final String label) {
            this.buttonLabel = label;
            return this;
        }

        @Override
        public BackEndActionDto withButtonTooltip(final String tooltip) {
            this.buttonTooltip = tooltip;
            return this;
        }

        @Override
        public BackEndActionDto withSuccessMessage(final String text) {
            this.successMessage = text;
            return this;
        }

        @Override
        public BackEndActionDto withFailureMessage(final String text) {
            this.failureMessage = text;
            return this;
        }

    }

}
